/**
  *  \file arch/unix/platform.cc
  *  \brief Platform-specific things for Unix
  */

#include <cstdlib>
#include <iostream>
#include <sys/stat.h>
#include <unistd.h>
#include <locale.h>
#include <wchar.h>
#include <stddef.h>
#include <string.h>
#include <limits.h>
#include "arch/platform.h"      // not arch/unix/platform.h, we need PCC2_xxxDIR
#include "defs.h"
#include "version.h"
#include "cpluslib/utf8.h"
#include "util/console.h"
#include "io/dirs.h"

const char Platform::LINE_SEPARATOR[] = "\n";
const char Platform::PATH_SEPARATOR;

void
Platform::postGraphicsInit()
{
    /* when running as root, (attempt to) drop privileges. */
    if (geteuid() == 0)
        setuid(getuid());
}

void
Platform::initPlatform()
{
    // Figure out user's file name character set
    setlocale(LC_CTYPE, "");
    setlocale(LC_MESSAGES, "");
}

/****************************** Directories ******************************/

/** true iff directory fn exists. */
static bool
exists(const char* fn)
{
    struct stat st;
    return stat(fn, &st) == 0 && S_ISDIR(st.st_mode);
}

/** Get data directory, helper. Data directory is either
    $pdatadir or $pdatadir/version. */
static string_t
getDataDirectoryH()
{
    if (exists(PCC2_DATADIR "/" PCC2_VERSION))
        return string_t(PCC2_DATADIR "/" PCC2_VERSION);
    else
        return string_t(PCC2_DATADIR);
}

/** Get data directory (planets root). */
string_t
Platform::getDataDirectory()
{
#ifdef HAVE_INPLACE_ENABLED
    if (exists(PCC2_SRCDIR "/specs"))
        return string_t(PCC2_SRCDIR "/specs");
#endif
    return getDataDirectoryH();
}

/** Get resource directory. */
string_t
Platform::getResourceDirectory()
{
#ifdef HAVE_INPLACE_ENABLED
    if (exists(PCC2_SRCDIR "/resource"))
        return string_t(PCC2_SRCDIR "/resource");
#endif
    return getDataDirectoryH() + "/resource";
}

string_t
Platform::getLanguageCode()
{
    if (const char* c = std::getenv("LC_ALL"))
        return c;
    if (const char* c = std::getenv("LC_MESSAGES"))
        return c;
    if (const char* c = std::getenv("LANG"))
        return c;
    return string_t();
}

string_t
Platform::getUserSettingsDirectory(int version)
{
    /* Settings are stored in "~/.pcc2" (new version), or directly in
       $HOME (old version) */
    const char* c = std::getenv("HOME");
    if (!c)
        c = "/";

    if (version > 0)
        return Directory::makeFileName(c, ".pcc2");
    else
        return c;
}


/************************** Character Conversion *************************/

/*
 *  General principles:
 *
 *  We know that our UTF-8 coder can handle all Unicode characters
 *  from the BMP, and we assume that wchar_t actually is Unicode.
 *
 *  File name character set and command-line character set are the
 *  same, namely the libc multibyte character set.
 *
 *  It is possible that some file names do not decode correctly as
 *  multibyte characters, for example, files written with Latin-1
 *  locale and loaded in UTF-8 locale. To handle these, we use a
 *  combination of error characters and ASCII, assuming that the
 *  multibyte character set actually is some sort of wellformed UTF-8
 *  (i.e. not a character set with "invisible" shift states, or
 *  non-minimal UTF-8).
 */


/** Convert external string to UTF-8.
    \param out [out] Receives the data
    \param data [in] Data
    \param size [in] Number of bytes in data
    \return Number of bytes processed */
static size_t
convertExternalToUtf8(string_t& out, const char* data, size_t size)
{
    /* Convert string */
    size_t total = 0;
    mbstate_t mbs;
    memset(&mbs, 0, sizeof(mbs));
    while (size > 0) {
        wchar_t wc;
        size_t processed = mbrtowc(&wc, data, size, &mbs);
        if (processed == 0) {
            /* success. This means mbrtowc found a null byte. */
            return total + size;
        } else if (processed > size || uint16_t(wc) != wc || Utf8Reader::isErrorCharacter(wc)) {
            /* error cases, (size_t)-1 [incomplete] or (size_t)-2 [bogus],
               or non-representable character */
            return total;
        } else {
            /* regular case */
            size -= processed;
            data += processed;
            total += processed;
            appendUtf8(out, wc);
        }
    }

    return total;
}

/** Convert argv array to UTF-8. */
char**
Platform::convertArgumentsToUtf8(char** argv)
{
    try {
        // Count arguments
        int nargs = 0;
        while (argv[nargs] != 0)
            ++nargs;

        // Allocate result buffer
        char** results(new char*[nargs+1]);
        for (int i = 0; i < nargs; ++i) {
            string_t tmp(convertFileNameToUtf8(argv[i]));
            results[i] = new char[tmp.size()+1];
            tmp.copy(results[i], tmp.size(), 0);
            results[i][tmp.size()] = 0;
        }
        results[nargs] = 0;

        return results;
    }
    catch (...) {
        // If we cannot convert the arguments, things are messed up so badly
        // that we cannot start. Terminate as quickly as possible.
        static const char msg[] = "Unable to initialize application.\n";
        ::write(2, msg, sizeof(msg)-1);
        std::exit(255);
    }
}

/** Release arguments. */
void
Platform::freeArguments(char** argv)
{
    for (int i = 0; argv[i] != 0; ++i)
        delete[] argv[i];
    delete[] argv;
}

/** Convert UTF-8 to filename string. */
bool
Platform::convertUtf8ToFileName(string_t& result, const string_t str, bool placeholders)
{
    mbstate_t mbs;
    memset(&mbs, 0, sizeof(mbs));

    Utf8Reader rdr(str.data(), str.size());
    while (rdr.hasMore()) {
        const uint32_t ch = rdr.get();
        if (Utf8Reader::isErrorCharacter(ch)) {
            /* It's an error character. Take over as-is. */
            result.append(1, char(ch - Utf8Reader::makeErrorCharacter(0x80) + 0x80));
        } else {
            /* Regular Unicode character */
            char tmp[MB_LEN_MAX];
            size_t produced = wcrtomb(tmp, ch, &mbs);
            if (produced > sizeof(tmp)) {
                /* error, we cannot encode this character. For example:
                   the user entered a character which the current locale
                   cannot represent */
                if (placeholders) {
                    tmp[0] = '?';
                    produced = 1;
                } else {
                    return false;
                }
            }
            result.append(tmp, produced);
        }
    }
    return true;
}

/** Convert filename string to UTF-8. */
string_t
Platform::convertFileNameToUtf8(const string_t fn)
{
    string_t result;
    size_t converted = convertExternalToUtf8(result, fn.data(), fn.size());
    while (converted < fn.size()) {
        uint8_t ch = fn[converted++];
        if (ch < 0x80)
            result += char(ch);
        else
            appendUtf8(result, Utf8Reader::makeErrorCharacter(ch));
    }
    return result;
}

/******************************** Console ********************************/

/*
 *  Console behaviour: for simplicity, let's start with using the same
 *  coding routines as for file names. This means, invalid input
 *  survives a pass through us in the same way as it survives in a
 *  file name, namely often but not always.
 *
 *  For output, unencodable characters are replaced by '?', so people
 *  at least see something when output is produced in a totally
 *  foreign language.
 */

static void
unixConsoleHandler(const Console& con, int n)
{
    while (n--) {
        const Console::Message& m = con.getLine(n);
        string_t conv;
        Platform::convertUtf8ToFileName(conv, m.text, true);
        if (m.mc >= LOG_WARN)
            std::cerr << conv << std::endl;
        else
            std::cout << conv << std::endl;
    }
}

void
Platform::initConsoleOutput()
{
    console.console_changed.add(unixConsoleHandler);
}

bool
Platform::readConsoleLine(string_t prompt, string_t& line)
{
    // Show prompt
    string_t convPrompt;
    Platform::convertUtf8ToFileName(convPrompt, prompt, true);
    std::cout << convPrompt << std::flush;

    // Read line
    string_t userLine;
    if (!getline(std::cin, userLine))
        return false;
    line = convertFileNameToUtf8(userLine);
    return true;
}

void
Platform::showOutput()
{
    /* Note that although this function is empty for Unix, it must be
       implemented here, not as an inline function in platform.h. 
       Otherwise, it will conflict with the required non-inline
       implementation in pdk-platform.cc, which doesn't have its own
       header file. */
}
