/**
  *  \file game/load-dir.cc
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "io/io.h"
#include "game/global.h"
#include "game/load.h"
#include "game/struct.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "game/utilparse.h"
#include "game/stat.h"
#include "game/racename.h"
#include "io/dirs.h"
#include "game/classicvcr.h"

namespace {
    class OutboxReader : public GOutboxReader {
        GGameTurn& trn;
        int player;
     public:
        OutboxReader(GGameTurn& trn, int player)
            : trn(trn), player(player)
            { }
        void addMessage(string_t msg, GPlayerSet receivers)
            {
                addMessageFromFile(trn, msg, player, receivers);
            }
    };
}

/** Load Version-3.5 Message Outbox File (mess35x.dat).
    \param trn   Game turn record
    \param s     Stream, read pointer at position 0
    \param player which player's outbox to load */
static inline void
loadOutbox35(GGameTurn& trn, Stream& s, int player)
{
    trn.getOutbox(player).setFormat(GOutbox::fWinplan);
    OutboxReader(trn, player).loadOutbox35(s);
}

/** Load version 3.0 (Dosplan) outbox file (messX.dat).
    \param trn    target turn object
    \param s      stream, pointer at start of directory
    \param player which player's outbox to load */
static inline void
loadOutbox(GGameTurn& trn, Stream& s, int player)
{
    trn.getOutbox(player).setFormat(GOutbox::fDos);
    OutboxReader(trn, player).loadOutbox(s);
}

static void
loadKore(GGameTurn& trn, Stream& s, int player)
{
    char buffer[50 * 12];

    /* Header */
    int header_size = s.read(buffer, 102);
    if (header_size != 102) {
        /* Some programs generate 0-length koreX.dat files. Ignore those. */
        if (header_size == 0) {
            console.write(LOG_INFO, format(_("File `%s' is empty and will be ignored."),
                                           s.getName()));
            return;
        } else {
            throw FileTooShortException(s);
        }
    }
    if (getInt16(buffer, 0) != trn.getTurnNumber()) {
        console.write(LOG_INFO, format(_("File `%s' is stale and will be ignored."),
                                       s.getName()));
        return;
    }

    /* Minefields */
    loadKoreMinefields(trn, s, 500, player);

    /* Ion storms */
    loadKoreIonStorms(trn, s, 50);

    /* Explosions */
    loadKoreExplosions(trn, s, 50);

    /* Ufos */
    loadUfos(trn, s, 1, 100);

    /* Visual contacts */
    if (s.read(buffer, 24) == 24 && std::memcmp(buffer, "1120", 4) == 0) {
        int32 count = getInt32(buffer, 20);
        if (count < 0 || count > NUM_SHIPS)
            throw FileFormatException(s, _("Unbelievable number of visual contacts"));
        loadTargets(trn, s, count, TargetEncrypted, LoadCurrent, GPlayerSet(player));
    }
}

/** Load SKOREx.DAT file. */
static void
loadSkore(GGameTurn& trn, Stream& s, int /*player*/)
{
    using std::memcmp;

    /* Read header. It is not fatal if we cannot read it; some programs
       (Winplan?) generate blank SKORE.DAT files. */
    TSkoreHeader hdr;
    if (!getStructure(s, hdr))
        return;

    /* Do we have extended Ufos? */
    if (memcmp(hdr.signature, "yAmsz", 5) == 0 && hdr.rst_version > 0 && hdr.num_ufos > 100)
        loadUfos(trn, s, 101, hdr.num_ufos - 100);
}

/** Load unpacked game directory. */
void
loadDirectory(GGameTurn& trn, Directory& d, int player)
{
    TInt16 n;
    TGen   gen;

    const GPlayerSet src(player);

    console.write(LOG_INFO, format(_("Loading %s data..."), player_racenames.getAdjName(player)));

    /* gen.dat */
    Ptr<Stream> s = d.openFile(format("gen%d.dat", player), Stream::C_READ);
    getStructureT(*s, gen);
    if (gen.player_id != player)
        throw FileFormatException(*s, format(_("File is owned by player %d, should be %d"), gen.player_id, player));
    trn.setHaveData(player);
    trn.setGen(player, GGen(gen));

    getStatistics().mergeGen(trn.getGen(player));

    /* ships */
    s = d.openFile(format("ship%d.dat", player), Stream::C_READ);
    getStructureT(*s, n);
    loadShips(trn, *s, n.value, LoadCurrent, src);
    s = d.openFile(format("ship%d.dis", player), Stream::C_READ);
    getStructureT(*s, n);
    loadShips(trn, *s, n.value, LoadPrevious, src);

    /* targets */
    s = d.openFile(format("target%d.dat", player), Stream::C_READ);
    getStructureT(*s, n);
    loadTargets(trn, *s, n.value, TargetPlaintext, LoadCurrent, src);

    /* planets */
    s = d.openFile(format("pdata%d.dat", player), Stream::C_READ);
    getStructureT(*s, n);
    loadPlanets(trn, *s, n.value, LoadCurrent, src);
    s = d.openFile(format("pdata%d.dis", player), Stream::C_READ);
    getStructureT(*s, n);
    loadPlanets(trn, *s, n.value, LoadPrevious, src);

    /* starbases */
    s = d.openFile(format("bdata%d.dat", player), Stream::C_READ);
    getStructureT(*s, n);
    loadBases(trn, *s, n.value, LoadCurrent, src);
    s = d.openFile(format("bdata%d.dis", player), Stream::C_READ);
    getStructureT(*s, n);
    loadBases(trn, *s, n.value, LoadPrevious, src);

    /* messages */
    s = d.openFile(format("mdata%d.dat", player), Stream::C_READ);
    trn.getInbox().loadInbox(*s, player);

    /* shipxy */
    s = d.openFile(format("shipxy%d.dat", player), Stream::C_READ);
    loadShipXY(trn, *s, s->getSize(), LoadBoth, src, GPlayerSet());

    /* VCRs */
    s = d.openFile(format("vcr%d.dat", player), Stream::C_READ);
    trn.setVcrs(new GClassicVcrDatabase(*s));

    /* Outbox */
    s = d.openFileNT(format("mess35%d.dat", player), Stream::C_READ);
    if (s) {
        loadOutbox35(trn, *s, player);
    } else {
        s = d.openFileNT(format("mess%d.dat", player), Stream::C_READ);
        if (s)
            loadOutbox(trn, *s, player);
    }

    /* Commands */
    s = d.openFileNT(format("cmd%d.txt", player), Stream::C_READ);
    if (s) {
        trn.getCommands(player).loadCommandFile(*s, trn.getGen(player).getTimestamp());
    }

    /* Kore */
    s = d.openFileNT(format("kore%d.dat", player), Stream::C_READ);
    if (s)
        loadKore(trn, *s, player);

    /* Skore */
    s = d.openFileNT(format("skore%d.dat", player), Stream::C_READ);
    if (s)
        loadSkore(trn, *s, player);

    /* Util */
    s = d.openFileNT(format("util%d.dat", player), Stream::C_READ);
    if (s)
        GUtilParser(trn, player).read(*s);

    /* FLAK */
    maybeLoadFlakVcrs(trn, d, player);
}
