/**
  *  \file game/save-dir.cc
  *  \brief Saving Game Data -- Normal unpacked directories
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/checksum.h"
#include "cpluslib/format.h"
#include "io/streams.h"
#include "game/spec.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/struct.h"
#include "util/misc.h"
#include "cpluslib/misc.h"
#include "game/greg.h"
#include "game/racename.h"
#include "util/except.h"
#include "util/recode.h"
#include "arch/platform.h"
#include "io/dirs.h"

/** Send commands. This appends commands to be sent through the message file
    to the specified GOutbox, and stores the others into the cmdX.txt file
    in the game directory. You must then save he GOutbox.
    \param outbox [in/out] Command messages appended here
    \param cms    [in] Commands
    \param dir    [in] Game directory in which cmdX.txt is created
    \param gen    [in] Game directory control file */
static void
sendCommands(GOutbox& outbox, const GCommandContainer& cms, Directory& dir, const GGen& gen)
{
    Ptr<Stream> cmdfile;
    string_t accum;
    for (GCommandContainer::const_iterator i = cms.begin(); i != cms.end(); ++i) {
        string_t text = (*i)->getCommandText();
        if (text.length()) {
            if (text[0] != '$') {
                // Send through message file
                if (accum.length() + text.length() > 500) {
                    outbox.addMessage(accum, GPlayerSet(outbox.getSender()));
                    accum.clear();
                }
                if (accum.length())
                    accum += '\n';
                accum += text;
            } else {
                // Send through command file
                if (!cmdfile) {
                    // open file first, and write headers
                    cmdfile = dir.openFile(format("cmd%d.txt", gen.getPlayerId()), Stream::C_CREATE | Stream::C_WRITE);

                    string_t line =  _("# Additional commands");
                    line += Platform::LINE_SEPARATOR;
                    line += "$time ";
                    line += gen.getTimestamp().getTimestampAsString();
                    line += Platform::LINE_SEPARATOR;
                    line = convertUtf8ToGame(line);
                    cmdfile->writeT(line.data(), line.size());
                }
                text += Platform::LINE_SEPARATOR;
                text = convertUtf8ToGame(text);
                cmdfile->writeT(text.data(), text.size());
            }
        }
    }
    if (accum.length())
        outbox.addMessage(accum, GPlayerSet(outbox.getSender()));

    if (!cmdfile) {
        // Nothing written to command file, so erase it
        dir.eraseNT(format("cmd%d.txt", gen.getPlayerId()));
    }
}

/** Save game directory.
    \param trn         [in] Game turn object to save. This object may be temporarily
                       modified, hence it is not const, but you'll get it back in
                       the same state you gave it to me.
    \param dir         [in] Directory to save in
    \param player_mask [in] Set of players to save */
void
saveDirectory(GGameTurn& trn, Directory& dir, const GPlayerSet player_mask)
{
    bool have_control = false;
    bool have_fizz = false;
    int  control_id   = 0;

    try {
        loadFizz(dir);
        have_fizz = true;
    }
    catch (FileProblemException& e) {
        console.write(LOG_WARN, _("File \"fizz.bin\" not found. Game data may not be usable with other programs."));
    }
    for (int pid = 1; pid <= NUM_PLAYERS; ++pid) {
        if (!player_mask.contains(pid))
            continue;

        if (!have_control) {
            // FIXME: this assumes dir == global game directory
            control_id = loadControl(dir, pid);
            have_control = true;
        }

        console.write(LOG_INFO, format(_("Writing %s data..."), player_racenames.getAdjName(pid)));

        {
            Ptr<Stream> s;

            GGen& g = trn.getGen(pid);
            uint32_t sig_check = getChecksum(g.getSig1(), 10) + getChecksum(g.getSig2(), 10);

            uint32_t ship_chk = 0, planet_chk = 0, base_chk = 0;

            s = dir.openFile(format("ship%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            saveShips(trn.getCurrentUniverse(), *s, GPlayerSet(pid), ship_chk, save_ChecksumCount);
            s->writeT(trn.getGen(pid).getSig2(), 10);

            s = dir.openFile(format("pdata%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            savePlanets(trn.getCurrentUniverse(), *s, GPlayerSet(pid), planet_chk, save_ChecksumCount);
            s->writeT(trn.getGen(pid).getSig2(), 10);

            s = dir.openFile(format("bdata%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            saveBases(trn.getCurrentUniverse(), *s, GPlayerSet(pid), base_chk, save_ChecksumCount);
            s->writeT(trn.getGen(pid).getSig2(), 10);

            /* messages and commands. We add the commands to the message
               box, save that, and remove them again. */
            GOutbox& out = trn.getOutbox(pid);
            int marker = out.getCount();
            sendCommands(out, trn.getCommands(pid), dir, g);
            if (out.getFormat() == GOutbox::fWinplan)
                saveOutbox35(out, dir, pid);
            else
                saveOutbox(out, dir, pid);
            out.deleteMessagesAfter(marker);

            /* compute DIS checksums */
            ship_chk += computeShipChecksum(trn.getPreviousUniverse(), GPlayerSet(pid), save_ChecksumCount);
            planet_chk += computePlanetChecksum(trn.getPreviousUniverse(), GPlayerSet(pid), save_ChecksumCount);
            base_chk += computeBaseChecksum(trn.getPreviousUniverse(), GPlayerSet(pid), save_ChecksumCount);
            
            g.setChecksum(ShipSection,       ship_chk   + sig_check);
            g.setChecksum(PlanetSection,     planet_chk + sig_check);
            g.setChecksum(BaseSection,       base_chk   + sig_check);
            setFizzEntry(ShipSection,   pid, ship_chk   + sig_check);
            setFizzEntry(PlanetSection, pid, planet_chk + sig_check);
            setFizzEntry(BaseSection,   pid, base_chk   + sig_check);
            s = dir.openFile(format("gen%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            storeStructureT(*s, g.getGenRecord());

            updateGameRegistry(g.getTimestamp());
        }

        if (control_id > 0) {
            // we are playing DOS planets
            saveControl(dir, control_id);
            have_control = false;
        }
    }

    if (have_control)
        saveControl(dir, control_id);
    doneControl();
    if (have_fizz)
        saveFizz(dir);
}
