/**
  *  \file int/args.h
  *  \brief Interpreter: Argument Handling Utilities
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_ARGS_H
#define PCC_V2_INT_ARGS_H

#include "cpluslib/types.h"

class IntDataSegment;
class IntValue;

/** Arguments to a function call. This describes a slice of an
    IntDataSegment used as parameters to a function call or array
    indexing operation.

    Its main use is to pack all necessary information conveniently
    together so we don't have to pass around three args all the
    time, and it provides useful functions to query the arguments. */
class IntArgBlock {
    const IntDataSegment& dseg;
    uint32_t index;
    uint32_t nargs;
 public:
    IntArgBlock(const IntDataSegment& dseg, uint32_t index, uint32_t nargs);
    IntValue* getNext();
    uint32_t getNumArgs() const;

    void checkArgumentCount(uint32_t need);
    void checkArgumentCount(uint32_t min, uint32_t max);
    void checkArgumentCountAtLeast(uint32_t min);
};

void checkArgumentCount(uint32_t have, uint32_t min, uint32_t max);


/**************************** Inline Functions ***************************/

/** Constructor.
    \param dseg  [in] Data segment
    \param index [in] Index of first argument
    \param nargs [in] Number of argument */
inline
IntArgBlock::IntArgBlock(const IntDataSegment& dseg, uint32_t index, uint32_t nargs)
    : dseg(dseg), index(index), nargs(nargs)
{ }

/** Get number of (remaining) arguments. Each getNext() will reduce
    this number. */
inline uint32_t
IntArgBlock::getNumArgs() const
{
    return nargs;
}     


#endif
