/**
  *  \file reshack/charname.cc
  *  \brief Resource Editor: Character Names
  *
  *  This module manages and uses Unicode character names. To use it,
  *  get the file NamesList.txt from Unicode, and place it in the
  *  resource directory as resource/reshack/charnames.txt.
  *
  *  Character names are then displayed along with the characters, and
  *  can be used to select characters.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <map>
#include <algorithm>
#include <vector>
#include <ctype.h>
#include "reshack/charname.h"
#include "resmgr/resmgr.h"
#include "io/textfile.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "cpluslib/utf8.h"
#include "arch/platform.h"

/** Global list of character names. Maps Unicode codepoints to names. */
static std::map<uint32_t, string_t> character_names;

/** Global list of character alias names. */
static std::vector<std::pair<string_t, string_t> > character_aliases;

/** Skip whitespace in a string.
    \param s [in] string
    \param n [in/out] current position. Updated to point to the first non-whitespace */
static void
skipWhitespace(const string_t& s, string_t::size_type& n)
{
    while (n < s.size() && isspace(uint8_t(s[n])))
        ++n;
}

/** Parse hex number. Reads hex digits from a string and assembles them into a number.
    \param s [in] string
    \param n [in/out] current position. Updated to point after the hex digits
    \param result [out] receives the hex number
    \return true iff any hex digits were read */
static bool
parseHex(const string_t& s, string_t::size_type& n, uint32_t& result)
{
    bool did = false;
    result = 0;
    while (n < s.size()) {
        char c = tolower(uint8_t(s[n]));
        if (c >= '0' && c <= '9') {
            result = 16*result + (c-'0');
            did = true;
            ++n;
        } else if (c >= 'a' && c <= 'f') {
            result = 16*result + 10 + (c-'a');
            did = true;
            ++n;
        } else {
            break;
        }
    }
    return did;
}

/** Given a character name, get its Id. */
static uint32_t
getCharacterByName(string_t name)
{
    for (std::map<uint32_t, string_t>::const_iterator it = character_names.begin(); it != character_names.end(); ++it)
        if (it->second == name)
            return it->first;
    return 0;
}

/**************************** Public Functions ***************************/

/** Load character name table. If the table cannot be loaded, nothing
    happens. The program continues without character names. */
void
loadCharacterNames()
{
    /* Default names for our characters */
    for (int i = 0; i < 16; ++i) {
        character_names[0xE100 + i] = format("[PCC2] REPLACEMENT UPPER LEFT %X", i);
        character_names[0xE110 + i] = format("[PCC2] REPLACEMENT UPPER RIGHT %X", i);
        character_names[0xE120 + i] = format("[PCC2] REPLACEMENT LOWER LEFT %X", i);
        character_names[0xE130 + i] = format("[PCC2] REPLACEMENT LOWER RIGHT %X", i);
    }
    character_names[0xE140] = "[PCC2] TOP ARC CLOCKWISE ARROW";
    character_names[0xE141] = "[PCC2] NESTED RECTANGLES ORNAMENT";
    character_names[0xE142] = "[PCC2] ORNAMENT LEFT";
    character_names[0xE143] = "[PCC2] ORNAMENT RIGHT";
    character_names[0xE144] = "[PCC2] THIN VERTICAL BAR";

    /* Read character name list */
    Ptr<Stream> s = Stream::openFileNT(Directory::makeFileName(Platform::getResourceDirectory(), "reshack/charnames.txt"), Stream::C_READ);
    if (s) {
        static const uint32_t none = 0xFFFFFFFF;
        TextFile tf(*s);
        string_t line;
        uint32_t current = none;
        while (tf.getLine(line)) {
            /* 001D    <control>
                       = INFORMATION SEPARATOR THREE
               0020    SPACE */
            uint32_t nchar;
            string_t::size_type n;

            /* Strip "comment" */
            n = line.find('(');
            if (n != line.npos) {
                while (n > 0 && line[n-1] == ' ')
                    --n;
                line.erase(n);
            }

            /* Parse it */
            n = 0;
            if (line.size() != 0 && parseHex(line, n, nchar)) {
                /* Line starts with a hex code */
                skipWhitespace(line, n);
                if (n < line.size() && line[n] != '<') {
                    /* It's a real name */
                    character_names[nchar] = strUCase(line.substr(n));
                    current = none;
                } else {
                    /* Not a real name. Note number for later */
                    current = nchar;
                }
            } else {
                /* Line does not start with a hex code */
                skipWhitespace(line, n);
                if (current != none && n < line.size() && line[n] == '=') {
                    /* Alternate name line for a character with bad name */
                    ++n;
                    skipWhitespace(line, n);
                    if (n < line.size()) {
                        character_names[current] = strUCase(line.substr(n));
                        current = none;
                    }
                }
            }
        }
    }

    /* Read character alias list */
    s = Stream::openFileNT(Directory::makeFileName(Platform::getResourceDirectory(), "reshack/alias.txt"), Stream::C_READ);
    if (s) {
        TextFile tf(*s);
        string_t line;
        while (tf.getLine(line)) {
            string_t::size_type n = line.find('=');
            if (n != line.npos) {
                character_aliases.push_back(std::make_pair(strTrim(line.substr(0, n)),
                                                           strTrim(line.substr(n+1))));
            }
        }
    }
}

/** Get name for a Unicode character.
    \param ch Unicode codepoint */
string_t
getCharacterName(uint32_t ch)
{
    string_t result = format("U+%04X", ch);

    std::map<uint32_t, string_t>::iterator i = character_names.find(ch);
    if (i != character_names.end()) {
        /* Name known from table */
        result += ' ';
        result += i->second;
    } else if (ch > ' ' && ch < 127) {
        /* No name table loaded, but it's an ASCII character */
        result += " ";
        result += char(ch);
    }

    return result;
}

static bool
compareValues(const std::pair<uint32_t, uint32_t>& a,
              const std::pair<uint32_t, uint32_t>& b)
{
    return a.second > b.second
        || (a.second == b.second
            && a.first < b.first);
}

/** Enumerate all characters that match a given expression. This
    checks for several criteria at once, and sorts the most likely
    results at the beginning:
    - character numbers (hex)
    - single characters
    - substring matches, where a match at the beginning of a word
      is worth more than one inside a word, and a match immediately
      after LETTER or DIGIT is worth most (i.e. "OM" matches first
      all the "LETTER OMEGA"s, then the "COMMA".

    \param expr  [in] Expression to search for
    \param limit [in] Maximum character code to return (i.e. this will
                 not return characters whose code is above this
                 even if they match)
    \param out   [out] Generate output here */
void
matchCharacterNames(string_t expr, const uint32_t limit, CharacterList& out)
{
    // Clear output. Quick exit if expression is empty
    out.clear();
    if (expr.size() == 0)
        return;

    // Collect characters and their values
    std::map<uint32_t, uint32_t> values;

    // Check for "U+nnnn" match
    if (expr.size() >= 2) {
        string_t::size_type n = 0;
        skipWhitespace(expr, n);
        if (expr.size() - n > 3 && (expr[n] == 'U' || expr[n] == 'u') && expr[n+1] == '+')
            n += 2;
        uint32_t result;
        if (parseHex(expr, n, result) && result <= limit) {
            skipWhitespace(expr, n);
            if (n == expr.size()) {
                // Match
                values[result] += 1000;
            }
        }
    }

    // Check for single character match
    Utf8Reader rdr(expr);
    if (rdr.hasMore()) {
        uint32_t ch = rdr.get();
        if (!rdr.hasMore()) {
            values[ch] += 900;
        }
    }

    // Check for name match
    expr = strUCase(expr);
    for (std::map<uint32_t, string_t>::const_iterator i = character_names.begin(); i != character_names.end() && i->first <= limit; ++i) {
        if (i->second.find("LETTER " + expr) != string_t::npos) {
            values[i->first] += 800;
        } else if (i->second.find("DIGIT " + expr) != string_t::npos) {
            values[i->first] += 700;
        } else if (i->second.find(" " + expr) != string_t::npos) {
            values[i->first] += 600;
        } else {
            string_t::size_type n = i->second.find(expr);
            if (n == 0) {
                values[i->first] += 600;
            } else if (n != string_t::npos) {
                values[i->first] += 500;
            }
        }
    }

    // Generate output
    std::vector<std::pair<uint32_t,uint32_t> > tmp(values.begin(), values.end());
    std::sort(tmp.begin(), tmp.end(), compareValues);
    for (std::vector<std::pair<uint32_t,uint32_t> >::iterator i = tmp.begin(); i != tmp.end(); ++i)
        out.push_back(i->first);
}

/************************** RHCharacterGenerator *************************/

RHCharacterGenerator::~RHCharacterGenerator()
{ }

/** Attempt to generate a character. This calls check() with a list of
    characters, possibly multiple times, to generate a character by
    combining these characters. check() can return true if it can
    actually handle the combination; in this case, generate() will
    return true. If no combination is found or check() rejects them
    all, it will return false.

    \todo Right now, this does a syntactic search using character
    names and an ad-hoc "alias.txt" file. This problem can also
    partially be solved by using the ":" lines from the name list
    (i.e. "00C2 LATIN CAPITAL LETTER A WITH CIRCUMFLEX", described as
    ": 0041 0302"). It doesn't help for all cases, i.e. generating
    CYRILLIC A from LATIN A. */
bool
RHCharacterGenerator::generate(uint32_t id)
{
    uint32_t result[10];
    uint32_t rsize = 1;

    // Figure out character name. Only with the name, we can do something sensible
    std::map<uint32_t, string_t>::iterator i = character_names.find(id);
    if (i == character_names.end())
        return false;

    // Does it have an accent?
    string_t name = i->second;
    string_t::size_type n = name.find(" WITH ");
    if (n != string_t::npos) {
        uint32_t accent = getCharacterByName("COMBINING " + name.substr(n + 6));
        if (accent == 0) {
            accent = getCharacterByName("COMBINING " + name.substr(n + 6) + " ACCENT");
        }
        if (accent != 0) {
            result[1] = accent;
            rsize = 2;
            name.erase(n);
        }
    }

    // Check trivial character
    uint32_t basechar = getCharacterByName(name);
    if (basechar && rsize != 1) {
        result[0] = basechar;
        if (check(result, rsize))
            return true;
    }

    // Check aliases
    for (uint32_t index = 0; index < character_aliases.size(); ++index) {
        // Does it have a variant?
        string_t::size_type varpos = character_aliases[index].first.find('?');
        string_t::size_type reppos = character_aliases[index].second.find('?');
        const string_t& a = character_aliases[index].first;
        const string_t& b = character_aliases[index].second;
        if (varpos == string_t::npos || reppos == string_t::npos) {
            // Check direct alias
            if (name == a && (basechar = getCharacterByName(b)) != 0) {
                result[0] = basechar;
                if (check(result, rsize))
                    return true;
            }
        } else {
            // Check variations
            const char*const variations[] = {
                "CAPITAL LETTER",
                "SMALL LETTER"
            };
            for (uint32_t v = 0; v < countof(variations); ++v) {
                if (name == a.substr(0, varpos) + variations[v] + a.substr(varpos+1)
                    && (basechar = getCharacterByName(b.substr(0, reppos) + variations[v] + b.substr(reppos+1))) != 0)
                {
                    result[0] = basechar;
                    if (check(result, rsize))
                        return true;
                }
            }
        }
    }
    return false;
}

/**************************** RHCharacterName ****************************/

/** Constructor. */
RHCharacterName::RHCharacterName()
    : character(0)
{ }

void
RHCharacterName::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, 1);
    info.flags = li_FlexibleH;
}

void
RHCharacterName::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useColorScheme(getSkin());
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setTextAlign(0, 1);
    ctx.setSolidBackground();
    ctx.setColor(UIColor::tc_Static);

    outTextF(ctx, extent.x, extent.y + extent.h/2, extent.w, getCharacterName(character));

    can.update(extent);
}

/** Set character whose name is to be displayed.
    \param character Unicode codepoint */
void
RHCharacterName::setCharacter(uint32_t character)
{
    if (character != this->character) {
        this->character = character;
        drawWidget();
    }
}

/************************** RHCharacterNameList **************************/

/** Constructor.
    \param charlist list to display */
RHCharacterNameList::RHCharacterNameList(int id, CharacterList& charlist)
    : UIListbox(id, 15, 200, charlist.size()),
      charlist(charlist)
{
    update();
}

void
RHCharacterNameList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        int ele = getTop() + i;
        if (ele < int(charlist.size())) {
            outTextF(ctx, area.x, area.y, area.w, getCharacterName(charlist[ele]));
        }
    }
}

/** Update widget. Call this when the charlist has changed. */
void
RHCharacterNameList::update()
{
    setSize(charlist.size());
    adjustPosition();
    drawWidget();
}
